// PitIn PC Sketch by TRAINO
//
// based on Mr Watch(DCC Monitor) By yaasan
// Copyright(C) 2017- Desktop Station Co.,Ltd.
//
//  Ver 1 2021-1-16
//

#include "NmraDcc_v141.h"
#include <Wire.h>
#include "I2CLCDLib.h"
#include <EEPROM.h>

// definition,configuration
#define a 14 //PC0 
#define b 15 //PC1 
#define cc 16 //PC2 
#define d  17 //PC3
#define e  18 //PC4
#define f  19 //PC5
#define g  8 //PB0
#define dp 9 //PB1

#define  ca1 10 //pb2
#define  ca2 0 //pd0
#define  ca3 1 //pd1

#define red  4 // pd4
#define blu  5 // pd5
#define grn  6 // pd6

#define PIN_DCCIN 2
// INT0 <= U11
// INT1 <= U12

#define DECODER_ADDRESS 3
#define MAX_LOCTABLE 16
#define MAX_ACCTABLE 128

#define MODE_LOC 0
#define MODE_ACC 1
#define MODE_VMON 2
#define MODE_ERR 3
#define MODE_TOTAL 4

#define PULSE_MIN	40
#define PULSE_MAX	200

#define PULSE_BUF_MAX 128

/* DCC SpeedStep 28専用テーブル */
const uint8_t DCC_28SPEEDMAP[29] = {
  0b00000, 0b00010, 0b10010, 0b00011,
  0b10011, 0b00100, 0b10100, 0b00101,
  0b10101, 0b00110, 0b10110, 0b00111,
  0b10111, 0b01000, 0b11000, 0b01001,
  0b11001, 0b01010, 0b11010, 0b01011,
  0b11011, 0b01100, 0b11100, 0b01101,
  0b11101, 0b01110, 0b11110, 0b01111, 0b11111
};

//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

/* LCD functions */
I2CLCDLib LCD;

//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL1 = 0;


struct LocData {
  uint16_t	mAddr;
  uint16_t  mSpeed;
  uint16_t  mSpeedhex;  // ★ for hexadecimal 
  uint8_t	mDirection;
  uint32_t	mFunction;
};

unsigned char message[10]; //　★ DCC message

LocData gLocTable[MAX_LOCTABLE];
uint16_t gAccTable[MAX_ACCTABLE];

uint16_t gAccLatestAddr[4] = {0, 0, 0, 0};
uint16_t gAccLatestDir[4] = {0, 0, 0, 0};
uint16_t gAccLastAddr = 9999;


uint8_t gMode = MODE_LOC;
uint8_t gTopIndex_Loc = 0xFF;
uint8_t gTopIndex_Acc = 0;

uint8_t gWatchMode = 0; // Analogを初期モードにする

uint8_t gButtonStatus = 0;

uint16_t gCount_Err = 0;
uint16_t gCount_Idle = 0;
uint16_t gCount_Reset = 0;
uint16_t gCount_PreMin = 255;
uint16_t gCount_PreMax = 0;


void(* resetFunc) (void) = 0;  //declare reset function at address 0
void OutputResults();
void ClearTable();
void DisplayFunctionBit(uint32_t inFunctionBits);
void DisplayAccBit(uint16_t inFunctionBits);
uint16_t ReadEdcVoltage();

void updateDccSpeed(word inLocoAddr, uint16_t inAddr, uint8_t inSpeed, uint8_t inDir, uint8_t inSpeedSteps );
// void decodeLocoPacket(uint8_t inPacket1, uint8_t inPacket2); unused
uint16_t getLocAddrIndex(uint16_t inAddr);

unsigned long checkTime=0, edgeTime=0, oldEdgeTime=0,devTime=0; 
unsigned long scale; // ex 150
unsigned long alarm; // ex 150
unsigned int lowCutoff; 
unsigned long edgeCount = 0; // 最初のリクエストを入れるため
char validInt1 = 0;
char dbuff[20];
#define UPDATEMS 400 // 1000=>400

unsigned int displayValue=0;
unsigned char blink=0;
unsigned char updateReq=0;
byte mph=0;
void edge0() {   /* Edge detect for speedmeter*/
  validInt1 = 1;
  edgeTime = millis();
  edgeCount++;
  if(displayValue < 100) updateReq=1;
}
//
// speed calucration 
//  edge interrult
//  or UPDATEMS cycle
//
void calcSpeed() {

    //checkTime +=UPDATEMS;
    unsigned int cp;
    if(edgeCount!=0) {
      if (oldEdgeTime ==0)
         displayValue = 1;  // primary edge
      else // sedondary edge
      {     
        //speed = (1000000/35/scale/ 5.6  *314/(edgeTime-oldEdgeTime)*edgeCount);
        // ==>> speed = scale * 0.0056 * 3.14 * 3600  /(edgeTime-oldEdgeTime) * edgeCount
        cp =  edgeCount *  scale * 63 / (edgeTime - oldEdgeTime) ;
        if(cp <= 500) 
           displayValue = cp;
        else
           displayValue = 500;
      }
      oldEdgeTime = edgeTime;
      edgeCount = 0;
    }
    else // no edgeCount
    {
      if(oldEdgeTime==0) 
        displayValue = 0;
      else {
          if ((millis() - oldEdgeTime) > 2500){
            displayValue = 0;oldEdgeTime=0;
          } // 2500 => 4500 => 2500 ここは一度だけ通る 
          else {
            if (displayValue>2)
              displayValue =  1 *  scale * 32 / (millis() - oldEdgeTime) ; // 68/2 => 32
              
          }
      }
    }
    // detect alarm
    if((alarm != 0)&&(displayValue > alarm)) blink=1; else blink = 0;
    validInt1 = 0; // clear interrupt flag
}
      
//
// proguram starts here
//
void setup()
{
    //ClearTable();
    pinMode(ca1, OUTPUT); //7seg
    pinMode(ca2, OUTPUT); //7seg
    pinMode(ca3, OUTPUT);  //7seg
    pinMode(a, OUTPUT);  //7seg
    pinMode(b, OUTPUT);  //7seg
    pinMode(cc, OUTPUT);  //7seg
    pinMode(d, OUTPUT);  //7seg
    pinMode(e, OUTPUT);  //7seg
    pinMode(f, OUTPUT);  //7seg
    pinMode(g, OUTPUT);  //7seg
    pinMode(dp, OUTPUT);  //7seg
    digitalWrite(red,HIGH); // pull up the switch
    digitalWrite(blu,HIGH);
    digitalWrite(grn,HIGH);
    
  if (EEPROM.read( 0 ) == 255 ) EEPROM.write( 0, 0 ) ; // mode 1 = mph
  if (EEPROM.read( 1 ) == 255 ) EEPROM.write( 1, 150 ) ; // defaut gare = '1/150'
  if (EEPROM.read( 2 ) == 255 ) EEPROM.write( 2, 100 ) ; // alarm speed 
  if (EEPROM.read( 3 ) == 255 ) EEPROM.write( 3, 10 ) ; // low-cutoff value 

  mph =  EEPROM.read( 0 );
  scale = EEPROM.read( 1 );
  alarm = EEPROM.read( 2 );
  lowCutoff = EEPROM.read( 3 );

  // Configuration mode
  if (digitalRead(red) == LOW) {
    for (; digitalRead(red) == LOW;) {
       displayValue = 888;
          updateSeg();
        delay(5);    
     } // wait release the switch
   //  show version 
   for(unsigned char i=0;i<200;i++) {
          displayValue = 1;
          updateSeg();
        delay(5);    
     } // wait release the switch

    // Inquire model scale
    blink=1;
    for (;;) {
      displayValue = scale;
      updateSeg();
      delay(5);

      if (digitalRead(red) == LOW) {
        blink=0;
        for (; digitalRead(red) == LOW;){updateSeg();delay(5);};
        EEPROM.write(1, scale);
        break;
      }
      if (digitalRead(grn) == LOW) {
        for (; digitalRead(grn) == LOW;){updateSeg();delay(5);};
        switch (scale) {
          case 160: scale = 220; break;
          case 150: scale = 160; break;
          case 148: scale = 150; break;
          case 87: scale = 148; break;
          case 80: scale = 87; break;
          case 48: scale = 80; break;
          default: scale = 48; break; 
        }
      }
      if (digitalRead(blu) == LOW) {
        for (; digitalRead(blu) == LOW;){updateSeg();delay(5);};
        switch (scale) { 
          case 80: scale = 48; break;
          case 87: scale = 80; break;
          case 148: scale = 87; break;
          case 150: scale = 148; break;
          case 160: scale = 150; break;
          case 220: scale = 160; break;
          default: scale = 220; break;        
        }
      } 
    }
    // inquire alarm setting
    blink=1;
    for (displayValue = alarm;;) {
      updateSeg();
      delay(5);

      if (digitalRead(red) == LOW) {
        for (; digitalRead(red) == LOW;){updateSeg();delay(5);};
        alarm = displayValue;
        EEPROM.write(2, displayValue);
        blink=0;
        break;
      }
      if (digitalRead(grn) == LOW) {
        for (; digitalRead(grn) == LOW;){updateSeg();delay(5);};
        if(displayValue < 190) displayValue+=10;
        else displayValue=200;
      }
      if (digitalRead(blu) == LOW) {
        for (; digitalRead(blu) == LOW;){updateSeg();delay(5);};
        if(displayValue > 10) displayValue-=10;
        else displayValue=0;

      } 
    }
    // inquire low-cutoff value 
    blink=1;
    for (displayValue = lowCutoff;;) {
      updateSeg();
      delay(5);

      if (digitalRead(red) == LOW) {
        for (; digitalRead(red) == LOW;){updateSeg();delay(5);};
        lowCutoff=displayValue;
        EEPROM.write(3, displayValue);
        blink=0;
        break;
      }
      if (digitalRead(grn) == LOW) {
        for (; digitalRead(grn) == LOW;){updateSeg();delay(5);};
        if(displayValue < 20) displayValue++;
        else displayValue=20;
      }
      if (digitalRead(blu) == LOW) {
        for (; digitalRead(blu) == LOW;){updateSeg();delay(5);};
        if(displayValue > 1) displayValue--;
        else displayValue=0;
      } 
    }
  }
  else {  // show scale 
    for(unsigned char i=0;i<200;i++) {
         displayValue = scale;
            updateSeg();
          delay(5);    
    }
  }
  // End of conf.

  //Reset task
  gPreviousL5 = millis();
  attachInterrupt(0, edge0, FALLING); 

  displayValue=0;
  checkTime  = millis();
  
}

void segment(char c) {
  switch (c) {
    case '-':    //digitalWrite(a, HIGH); digitalWrite(b, HIGH);  digitalWrite(cc, HIGH);digitalWrite(d,HIGH); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC |= _BV(PC0);    PORTC |= _BV(PC1);      PORTC |= _BV(PC2);     PORTC |= _BV(PC3);    PORTC |= _BV(PC4);      PORTC |= _BV(PC5);    PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case ' ':    //digitalWrite(a, HIGH); digitalWrite(b, HIGH);  digitalWrite(cc, HIGH);digitalWrite(d,HIGH); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, HIGH);digitalWrite(dp, HIGH);break;
                    PORTC |= _BV(PC0);    PORTC |= _BV(PC1);      PORTC |= _BV(PC2);     PORTC |= _BV(PC3);    PORTC |= _BV(PC4);      PORTC |= _BV(PC5);    PORTB |= _BV(PB0);     PORTB |= _BV(PB1);  break;     
    case '0':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d,LOW); digitalWrite(e, LOW);  digitalWrite(f, LOW);digitalWrite(g, HIGH);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC &= ~_BV(PC5);   PORTB |= _BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '1':     //digitalWrite(a, HIGH); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, HIGH); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, HIGH);digitalWrite(dp, HIGH);break;
                    PORTC |= _BV(PC0);    PORTC &= ~_BV(PC1);      PORTC &= ~_BV(PC2);     PORTC |= _BV(PC3);    PORTC |= _BV(PC4);      PORTC |= _BV(PC5);    PORTB |= _BV(PB0);     PORTB |= _BV(PB1);  break;     
    case '2':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, HIGH);digitalWrite(d, LOW); digitalWrite(e, LOW);  digitalWrite(f, HIGH);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);   PORTC |= _BV(PC2);    PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC |= _BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '3':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC |= _BV(PC4);    PORTC |= _BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '4':    //digitalWrite(a, HIGH); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, HIGH); digitalWrite(e, HIGH);  digitalWrite(f, LOW);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC|= _BV(PC0);   PORTC &= ~_BV(PC1);   PORTC &= ~_BV(PC2);    PORTC |= _BV(PC3);     PORTC |= _BV(PC4);    PORTC &= ~_BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '5':     //digitalWrite(a, LOW); digitalWrite(b, HIGH);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, HIGH);  digitalWrite(f, LOW);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC |= _BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC |= _BV(PC4);    PORTC &= ~_BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '6':     //digitalWrite(a, LOW); digitalWrite(b, HIGH);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, LOW);  digitalWrite(f, LOW);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC |= _BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC &= ~_BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '7':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, HIGH); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, HIGH);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);      PORTC &= ~_BV(PC2);     PORTC |= _BV(PC3);    PORTC |= _BV(PC4);      PORTC |= _BV(PC5);    PORTB |= _BV(PB0);     PORTB |= _BV(PB1);  break;     
    case '8':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, LOW);  digitalWrite(f, LOW);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC &= ~_BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '9':     //digitalWrite(a, LOW); digitalWrite(b, LOW);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, HIGH);  digitalWrite(f, LOW);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC &= ~_BV(PC0);   PORTC &= ~_BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC |= _BV(PC4);    PORTC &= ~_BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case 'L':    //digitalWrite(a, HIGH); digitalWrite(b, HIGH);  digitalWrite(cc, HIGH);digitalWrite(d, LOW); digitalWrite(e, LOW);  digitalWrite(f, LOW);digitalWrite(g, HIGH);digitalWrite(dp, HIGH);break;
                    PORTC |= _BV(PC0);    PORTC |= _BV(PC1);   PORTC |= _BV(PC2);     PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC &= ~_BV(PC5);   PORTB |= _BV(PB0);     PORTB |= _BV(PB1);  break; 
    case 'o':    //digitalWrite(a, HIGH); digitalWrite(b, HIGH);  digitalWrite(cc, LOW);digitalWrite(d, LOW); digitalWrite(e, LOW);  digitalWrite(f, HIGH);digitalWrite(g, LOW);digitalWrite(dp, HIGH);break;
                    PORTC |= _BV(PC0);   PORTC |= _BV(PC1);   PORTC &= ~_BV(PC2);    PORTC &= ~_BV(PC3);  PORTC &= ~_BV(PC4);    PORTC |= _BV(PC5);   PORTB &= ~_BV(PB0);     PORTB |= _BV(PB1);  break; 
    case '.' :    // vvvvvvvvvvv continue
    default :     //digitalWrite(a, HIGH); digitalWrite(b, HIGH);  digitalWrite(cc, HIGH);digitalWrite(d,HIGH); digitalWrite(e, HIGH);  digitalWrite(f, HIGH);digitalWrite(g, HIGH);digitalWrite(dp, LOW);break;
                   PORTC |= _BV(PC0);    PORTC |= _BV(PC1);      PORTC |= _BV(PC2);     PORTC |= _BV(PC3);    PORTC |= _BV(PC4);      PORTC |= _BV(PC5);    PORTB |= _BV(PB0);     PORTB &= ~_BV(PB1);  break;     

  }
}

unsigned char clm=0;

void updateSeg() {
  if(blink) {
    if(millis()%1000 > 600) {
      digitalWrite(ca1, LOW); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
      return;
    }
  }
  switch (clm=(++clm)%3) {            
    case 0:     digitalWrite(ca1, LOW); digitalWrite(ca2, LOW);  digitalWrite(ca3, HIGH);
                if(displayValue<100)
                  segment(' ');
                else
                  segment(displayValue/100+'0');
                break;
    case 1:     digitalWrite(ca1, LOW); digitalWrite(ca2, HIGH);  digitalWrite(ca3, LOW);
                if(displayValue<10)
                  segment(' ');
                else
                  segment(displayValue/10%10+'0');
                break;
    case 2:     digitalWrite(ca1, HIGH); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
                segment(displayValue%10+'0');
                break;

  }
}
void updateSegLC() { // with low-cutoff function
  unsigned int dv;
  if(mph) dv = displayValue*6/10; else dv = displayValue;
  
  // scan button
  if((digitalRead(blu) == LOW) && (mph!=0)) { mph = 0;EEPROM.write(0,0);}  
  if((digitalRead(grn) == LOW) && (mph==0)) { mph = 1;EEPROM.write(0,1);}  
  if(blink) {
    if(millis()%1000 > 600) {
      digitalWrite(ca1, LOW); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
      return;
    }
  }
  if ((lowCutoff!=0)&&(displayValue<=lowCutoff)) {
    switch(clm=(++clm)%4) {
    case 3:     if(mph) {
                  digitalWrite(ca1, HIGH); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
                  segment('.');
                  break;
                }  
                clm=0;
                // k/m => continue case 0 
    case 0:     digitalWrite(ca1, LOW); digitalWrite(ca2, LOW);  digitalWrite(ca3, HIGH);
                segment('-');
                break;
    case 1:     digitalWrite(ca1, LOW); digitalWrite(ca2, HIGH);  digitalWrite(ca3, LOW);
                segment('L');
                break;
    case 2:     digitalWrite(ca1, HIGH); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
                segment('o');
                break;
    }
    return;
  }
  switch (clm=(++clm)%4) {
    case 3:     if(mph) {
                  digitalWrite(ca1, HIGH); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
                  segment('.');
                  break;
                }  
                clm=0;
                // k/m => continue case 0 
    case 0:     digitalWrite(ca1, LOW); digitalWrite(ca2, LOW);  digitalWrite(ca3, HIGH);
                if(dv<100)
                  segment(' ');
                else
                  segment(dv/100+'0');
                break;
    case 1:     digitalWrite(ca1, LOW); digitalWrite(ca2, HIGH);  digitalWrite(ca3, LOW);
                if(dv<10)
                  segment(' ');
                else
                  segment(dv/10%10+'0');
                break;
    case 2:     digitalWrite(ca1, HIGH); digitalWrite(ca2, LOW);  digitalWrite(ca3, LOW);
                segment(dv%10+'0');
                break;
  }
}


unsigned char seg=0;
//
//
//
void loop() {
  uint8_t aUpdate = 0;
  unsigned char btn = 0;
  unsigned char seg = 0;

  if ( (millis() - gPreviousL1) >= 5)  {
    //Reset task ★処理の最初に変更
    gPreviousL1 = millis();
    updateSegLC();  // disp one-digit 
  }

  if(updateReq) {
    calcSpeed();
    updateReq=0;
    gPreviousL5 = millis(); 
  }
  if ((millis() - gPreviousL5) >= 1000)   { //lowest update cycle (depend on stavirity of slow speed)
    calcSpeed();
    updateReq=0;
    gPreviousL5 = millis(); 
  }
}

//===========================================================================
void OutputResultsnotused()
{
  char aTexts[9];  // -> unsigned  ★計算にも使うため
  uint16_t aVolt;
  uint16_t aLocAddr;
  uint16_t aMax0 = 0;
  uint16_t aMax1 = 0;
  uint16_t aMin0 = 0xFFFF;
  uint16_t aMin1 = 0xFFFF;

  if ( gWatchMode == 0 )
  {
    if ( (gMode %= 2) == 0 ) // ★add PWM duty indicator
    {
      LCD.lcd_setCursor(0, 0);
      aVolt = ReadEdcVoltage();
      int dutyA = 0 ;
      int dutyB = 0 ;
      for (char c = 0; c < 100; c++) { // get 100 times
        if ( digitalRead(2) != 0 ) dutyA++;
        if ( digitalRead(8) != 0 ) dutyB++;
        delay(1);
      }
      if (dutyB > dutyA) dutyA = dutyB; // ★ show bigger PWM data
      sprintf(aTexts, "%2d.%dV%2d%%   ", aVolt / 10, aVolt % 10, dutyA);

      LCD.lcd_printStr(aTexts);
      LCD.lcd_setCursor(0, 1);
      if (validInt1)
        calcSpeed();
      else
        LCD.lcd_printStr("[Analog]");
      digitalWrite(7, LOW); delay(10); digitalWrite(7, HIGH); // assirt F.F. clear

    }
    else {
//      passingSpeedMeter();
    }

  }
  else
  {


    switch ( gMode %= 8 ) //** 5=>6=>8
    {

      case 0: // ★Main Display

        if ( gTopIndex_Loc != 0xFF)
        {
          /* ロコが登録中の時 */
          aLocAddr = (gLocTable[gTopIndex_Loc].mAddr);
          sprintf(aTexts, "L%3d%s%2d%%", aLocAddr, gLocTable[gTopIndex_Loc].mDirection == 0 ? "R" : "F", gLocTable[gTopIndex_Loc].mSpeed); // ★ホーマット変更
          LCD.lcd_setCursor(0, 0);
          LCD.lcd_printStr(aTexts);

          LCD.lcd_setCursor(0, 1);
          for ( int i = 0; i < 8; i++)
          {
            if ( ((gLocTable[gTopIndex_Loc].mFunction >> i ) & 1) == 0)
            {
              aTexts[i] = '-';
            }
            else
            {
              aTexts[i] = '0' + i; // ★ convert function num. to ascii code
            }
          }
          //LCD.lcd_printStr(aTexts);

          if (validInt1)
            calcSpeed();  // ★Add speed meter from roller/
          else
            LCD.lcd_printStr(aTexts);

        }
        else
        {
          /* ロコのパケットが流れてないとき */
          sprintf(aTexts, "Idle %3d", gCount_Idle);
          LCD.lcd_setCursor(0, 0);
          LCD.lcd_printStr(aTexts);
          LCD.lcd_setCursor(0, 1);
          LCD.lcd_printStr("  -DCC- ");

          /* アイドルパケットをリセット */
          gCount_Idle = 0;
        }

        break;


      case 1:
        //passingSpeedMeter();
        break;
      
      case 2:
        //Accessory
        for( int i = 0; i < 16; i++)
          {
          LCD.lcd_setCursor(i & 0b111, i >> 3);
          if( ((gAccTable[0] >> i ) & 1) == 0)
          {
            LCD.lcd_printStr("/");
          }
          else
          {
            LCD.lcd_printStr("|");
          }
          }
          break;

          


      case 3:
        //Accessory
        /*for( int i = 0; i < 16; i++)
          {
        	LCD.lcd_setCursor(i & 0b111, i >> 3);
        	if( ((gAccTable[0] >> i ) & 1) == 0)
        	{
        		LCD.lcd_printStr("/");
        	}
        	else
        	{
        		LCD.lcd_printStr("|");
        	}
          }*/
        sprintf(aTexts, "%3d%s%3d%s", gAccLatestAddr[0], (gAccLatestDir[0] == 0) ? "/" : "|", gAccLatestAddr[1], (gAccLatestDir[1] == 0) ? "/" : "|");
        LCD.lcd_setCursor(0, 0);
        LCD.lcd_printStr(aTexts);

        sprintf(aTexts, "%3d%s%3d%s", gAccLatestAddr[2], (gAccLatestDir[2] == 0) ? "/" : "|", gAccLatestAddr[3], (gAccLatestDir[3] == 0) ? "/" : "|");
        LCD.lcd_setCursor(0, 1);
        LCD.lcd_printStr(aTexts);


        break;

      case 4:
        //Edc
        aVolt = ReadEdcVoltage();
        sprintf(aTexts, "%2d.%dV   ", aVolt / 10, aVolt % 10);
        LCD.lcd_setCursor(0, 0);
        LCD.lcd_printStr(aTexts);
        LCD.lcd_setCursor(0, 1);
        LCD.lcd_printStr("  -DCC- ");
        break;

      case 5:

        Dcc.getPulseWidth(&aMax0, &aMin0, &aMax1, &aMin1);

        if ( (aMax0 != 0) && (aMin0 != 0x0FF))
        {

          sprintf(aTexts, "|%3d/%2d", aMin1, aMax1);
          LCD.lcd_setCursor(0, 0);
          LCD.lcd_printStr(aTexts);
        }

        if ( (aMax1 != 0) && (aMin1 != 0x0FF))
        {
          sprintf(aTexts, "O%3d/%2d", aMin0, aMax0);
          LCD.lcd_setCursor(0, 1);
          LCD.lcd_printStr(aTexts);
        }
        break;

      case 6:

        sprintf(aTexts, "Pre%2d/%2d", gCount_PreMin, gCount_PreMax);
        LCD.lcd_setCursor(0, 0);
        LCD.lcd_printStr(aTexts);

        sprintf(aTexts, "Err%5d", gCount_Err);
        LCD.lcd_setCursor(0, 1);
        LCD.lcd_printStr(aTexts);
        break;
        
      case 7:
          

        if ( gTopIndex_Loc != 0xFF)
        {
          sprintf(aTexts, "DCC Pckt");
          LCD.lcd_setCursor(0, 0);
          LCD.lcd_printStr(aTexts);

          unsigned char i = 0;
          unsigned char dd;
          for (i = 0; i < message[0]; i++) // ★表示文字の転送
          {
            dd = message[i + 1] >> 4;
            if (dd < 10) aTexts[2 * i] = dd + '0'; else aTexts[2 * i] = dd + 'A' - 10;
            dd = message[i + 1] % 16;
            if (dd < 10) aTexts[2 * i + 1] = dd + '0'; else aTexts[2 * i + 1] = dd + 'A' - 10;
          }
          for (; i <= 4; i++)
          {
            aTexts[2 * i] = ' ';
            aTexts[2 * i + 1] = ' ';
          }
          aTexts[9] = 0; // ★ termination=NULL
          LCD.lcd_setCursor(0, 1);
          LCD.lcd_printStr(aTexts);
        }
        else
        {
          /* ロコのパケットが流れてないとき */
          sprintf(aTexts, "Idle %3d", gCount_Idle);
          LCD.lcd_setCursor(0, 0);
          LCD.lcd_printStr(aTexts);
          LCD.lcd_setCursor(0, 1);
          LCD.lcd_printStr("  -DCC- ");

          /* アイドルパケットをリセット */
          gCount_Idle = 0;
        }

        break;
        

    }
  }
}

void ClearTable()
{
  for ( int i = 0; i < MAX_LOCTABLE; i++)
  {
    gLocTable[i].mAddr = 0;
    gLocTable[i].mSpeed = 0;
    gLocTable[i].mDirection = 0;
    gLocTable[i].mFunction = 0;
  }
}

void DisplayFunctionBit1(uint32_t inFunctionBits)
{

  for ( int i = 0; i < 29; i++)
  {
    if ( bitRead(inFunctionBits, i) == 0)
    {
      Serial.print(".");
    }
    else
    {
      Serial.print("O");
    }

  }
}

void DisplayAccBit(uint16_t inFunctionBits)
{

  for ( int i = 0; i < 16; i++)
  {
    if ( bitRead(inFunctionBits, i) == 0)
    {
      Serial.print(".");
    }
    else
    {
      Serial.print("O");
    }

  }
}

/* 全DCCメッセージを受信 */
void notifyDccMsg( DCC_MSG * Msg)
{
  uint8_t aErrorbyte = 0;
  uint16_t aLocoAddr = 0;
  uint16_t aAccAddr = 0;


  /* エラーチェック */
  for (uint8_t i = 0; i < Msg->Size; i++)
  {
    aErrorbyte = aErrorbyte ^ Msg->Data[i];
  }

  message[0] = Msg->Size;
  for (uint8_t i = 0; i < Msg->Size; i++)
  {
    message[i + 1] = Msg->Data[i];
  }


  if ( aErrorbyte != 0)
  {
    if ( gCount_Err < 65535)
    {
      gCount_Err++;
    }

    return;
  }
  else
  {
    gWatchMode = 1;
  }

  /* プリアンブル数計算 */
  if ( Msg->PreambleBits < gCount_PreMin)
  {
    gCount_PreMin = Msg->PreambleBits;
  }
  else if ( Msg->PreambleBits > gCount_PreMax)
  {
    gCount_PreMax = Msg->PreambleBits;
  }


  /* パケットチェック */
  switch (Msg->Data[0])
  {
    case 0x00:
      //Broadcast

      if ( Msg->Data[1] == 0x00)
      {
        gCount_Reset++;
      }
      else
      {
        /* その他のBroadcast処理 */
      }

      break;

    case 0xFF:
      //Idle packet
      gCount_Idle++;
      break;

    default:

      if ( Msg->Data[0] <= 127)
      {
        /* SHORTアドレスロコの操作 */
        aLocoAddr = Msg->Data[0];

        decodeLocoPacket(aLocoAddr, Msg->Data[1], Msg->Data[2]);


      }
      else if (( Msg->Data[0] >= 128) && ( Msg->Data[0] <= 191))
      {
        /* アクセサリ */
        aAccAddr = (((Msg->Data[0] & 0b111111) - 1) * 4) + ((7 - (Msg->Data[1] >> 4) & 0b111) << 8) +  (((Msg->Data[1] >> 1) & 0b11));


        if ( (Msg->Data[1] & 1) == 0)
        {
          gAccTable[ aAccAddr >> 4 ] = gAccTable[ aAccAddr >> 4 ] & (~(1 << (aAccAddr & 0xFFFF)) & 0xFFFF);


          if ( aAccAddr != gAccLastAddr)
          {
            gAccLatestAddr[3] = gAccLatestAddr[2];
            gAccLatestDir[3] = gAccLatestDir[2];

            gAccLatestAddr[2] = gAccLatestAddr[1];
            gAccLatestDir[2] = gAccLatestDir[1];

            gAccLatestAddr[1] = gAccLatestAddr[0];
            gAccLatestDir[1] = gAccLatestDir[0];

            gAccLatestAddr[0] = aAccAddr + 1;
            gAccLatestDir[0] = 0;
          }
          else
          {
            gAccLatestDir[0] = 0;
          }

          //uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is diverse.";
        }
        else
        {
          gAccTable[ aAccAddr >> 4 ] = gAccTable[ aAccAddr >> 4 ] | (1 << (aAccAddr & 0xFFFF)) ;
          //uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is straight.";

          if ( aAccAddr != gAccLastAddr)
          {
            gAccLatestAddr[3] = gAccLatestAddr[2];
            gAccLatestDir[3] = gAccLatestDir[2];

            gAccLatestAddr[2] = gAccLatestAddr[1];
            gAccLatestDir[2] = gAccLatestDir[1];

            gAccLatestAddr[1] = gAccLatestAddr[0];
            gAccLatestDir[1] = gAccLatestDir[0];

            gAccLatestAddr[0] = aAccAddr + 1;
            gAccLatestDir[0] = 1;
          }
          else
          {
            gAccLatestDir[0] = 1;
          }
        }

        gAccLastAddr = aAccAddr;


      }
      else if (( Msg->Data[0] >= 192) && ( Msg->Data[0] <= 231))
      {
        /* LONGアドレスロコの操作 */
        aLocoAddr = ((Msg->Data[0] - 192) << 8) + Msg->Data[1];

        decodeLocoPacket(aLocoAddr, Msg->Data[2], Msg->Data[3]);
      }
      else if (( Msg->Data[0] >= 232) && ( Msg->Data[0] <= 254))
      {
        /* 予約 */

      }



      break;

  }

}

void decodeLocoPacket(word inLocoAddr, uint8_t inPacket1, uint8_t inPacket2)
{
  uint8_t aSpeed;
  uint8_t aIndex = getLocAddrIndex(inLocoAddr);

  switch (inPacket1 >> 5)
  {
    case 0b000:
      break;

    case 0b001:

      switch (inPacket1 & 0b11111)
      {
        case 31:

          aSpeed = inPacket2 & 0b01111111;

          if ( aSpeed > 1)
          {
            aSpeed = aSpeed - 1;
          }

          /* リスト更新 */
          updateDccSpeed( inLocoAddr, aSpeed, (inPacket2 >> 7) & 1, 127 );
          break;
      }
      break;

    case 0b010:
      //REV

      aSpeed = inPacket1 & 0b11111;

      for (int i = 0; i < 29; i++)
      {
        if ( aSpeed == DCC_28SPEEDMAP[i])
        {
          /* リスト更新 */
          updateDccSpeed( inLocoAddr, i, 0, 29 );

          break;
        }
      }

      break;
    case 0b011:
      //FWD

      aSpeed = inPacket1 & 0b11111;

      for (int i = 0; i < 29; i++)
      {
        if ( aSpeed == DCC_28SPEEDMAP[i])
        {
          /* リスト更新 */
          updateDccSpeed( inLocoAddr, i, 1, 29 );

          break;
        }
      }

      break;

    case 0b100:
      bitWrite( gLocTable[aIndex].mFunction, 0, (inPacket1 >> 4) & 1);
      bitWrite( gLocTable[aIndex].mFunction, 1, (inPacket1 >> 0) & 1);
      bitWrite( gLocTable[aIndex].mFunction, 2, (inPacket1 >> 1) & 1);
      bitWrite( gLocTable[aIndex].mFunction, 3, (inPacket1 >> 2) & 1);
      bitWrite( gLocTable[aIndex].mFunction, 4, (inPacket1 >> 3) & 1);
      break;

    case 0b101:

      switch ((inPacket1 >> 4) & 1)
      {
        case 1:
          bitWrite( gLocTable[aIndex].mFunction, 5, (inPacket1 >> 0) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 6, (inPacket1 >> 1) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 7, (inPacket1 >> 2) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 8, (inPacket1 >> 3) & 1);
          break;
        case 0:
          bitWrite( gLocTable[aIndex].mFunction, 9, (inPacket1 >> 0) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 10, (inPacket1 >> 1) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 11, (inPacket1 >> 2) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 12, (inPacket1 >> 3) & 1);
          break;
      }
      break;

    case 0b110:


      switch (inPacket1 & 0b11111)
      {
        case 30:
          bitWrite( gLocTable[aIndex].mFunction, 13, (inPacket2 >> 0) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 14, (inPacket2 >> 1) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 15, (inPacket2 >> 2) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 16, (inPacket2 >> 3) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 17, (inPacket2 >> 4) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 18, (inPacket2 >> 5) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 19, (inPacket2 >> 6) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 20, (inPacket2 >> 7) & 1);
          break;

        case 31:
          bitWrite( gLocTable[aIndex].mFunction, 21, (inPacket2 >> 0) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 22, (inPacket2 >> 1) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 23, (inPacket2 >> 2) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 24, (inPacket2 >> 3) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 25, (inPacket2 >> 4) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 26, (inPacket2 >> 5) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 27, (inPacket2 >> 6) & 1);
          bitWrite( gLocTable[aIndex].mFunction, 28, (inPacket2 >> 7) & 1);
          break;
      }
  }



}

uint16_t getLocAddrIndex(uint16_t inAddr)
{
  uint16_t aIndex = 0xFFFF;

  for ( int i = 0; i < MAX_LOCTABLE; i++)
  {
    if ( (gLocTable[i].mAddr == inAddr) || (gLocTable[i].mAddr == 0) )
    {

      gLocTable[i].mAddr = inAddr;

      aIndex = i;

      break;
    }
  }

  return aIndex;
}

void updateDccSpeed( uint16_t inAddr, uint8_t inSpeed, uint8_t inDir, uint8_t inSpeedSteps )
{

  uint16_t aSpeed = 0, hSpeed = 0;

  for ( int i = 0; i < MAX_LOCTABLE; i++)
  {
    if ( (gLocTable[i].mAddr == inAddr) || (gLocTable[i].mAddr == 0) )
    {

      gLocTable[i].mAddr = inAddr;

      if ( inSpeed >= 1)
      {
        aSpeed = ((uint16_t)inSpeed * 100) / (uint16_t)(inSpeedSteps - 1); // ★　スピードステップ-1で計算する
        hSpeed = inSpeed;
      }
      else
      {
        hSpeed = aSpeed = 0;
      }


      if ( (aSpeed != gLocTable[i].mSpeed) || (gLocTable[i].mDirection != inDir))
      {
        gLocTable[i].mSpeed = aSpeed;
        gLocTable[i].mSpeedhex = hSpeed; // ★1
        gLocTable[i].mDirection = inDir;

        gTopIndex_Loc = i;

      }


      /* for文終了 */
      break;
    }


  }
}


uint16_t ReadEdcVoltage()
{

  //電圧換算
  // 1k / 11k  y=((x*11*5)/1/1024)
  uint16_t aVolt = analogRead(A2);

  //Limit
  if ( aVolt > 922)
  {
    //30Vで打ち止め
    aVolt = 922;
  }

  aVolt = aVolt >> 2;

  aVolt = (aVolt * 301) >> 7;

  if ( aVolt > 255)
  {
    //UCに入るサイズにリミット(25.5V)
    aVolt = 255;
  }

  return aVolt;

}
